/*
* @Author: Dimitrios Katsaros
* @Date:   2017-10-18 11:23:26
*/

/* An implementation of a multi dimentional array with either contiguous or
 * non contiguous memory.
 * The original goal of this is to be able to create arrays that can be applied
 * to a v4l2 device. The array has slowly permutted to being a multi purpose
 * structure, being widely used in different features of qteclib.
 * Operations on the array have not been designed to be
 * parallelizable, so take care of operations on the same memory region
 * when handing out the wrapper to separate threads.
 */

#ifndef __QTEC_LIB_CORE_ARRAY_H__
#define __QTEC_LIB_CORE_ARRAY_H__

#include <qtec/core/memory.h>
#include <qtec/core/iterator.h>

G_BEGIN_DECLS

#define QTEC_ARRAY_DIM_HEIGHT 0
#define QTEC_ARRAY_DIM_WIDTH 1
#define QTEC_ARRAY_DIM_DEPTH 2
#define QTEC_ARRAY_STRIDE_ROW 1
#define QTEC_ARRAY_STRIDE_DEPTH 2

typedef struct _QtecArray QtecArray;

typedef enum {
	QTEC_ARRAY_LE = 0,
	QTEC_ARRAY_BE = 1,
} QtecArrayEndianness;

#if G_BYTE_ORDER == G_LITTLE_ENDIAN
#define QTEC_ARRAY_HOST_ENDIANNESS QTEC_ARRAY_LE
#elif G_BYTE_ORDER == G_BIG_ENDIAN
#define QTEC_ARRAY_HOST_ENDIANNESS QTEC_ARRAY_BE
#else
#error unknown ENDIAN type
#endif

/* structure providing information on an array */
struct _QtecArrayInfo {
	/* the number of dimensions */
	guint dimno;
	/* A ppointer to an array of dimensions */
	guint *dims;
	/* A pointer to an array of strides */
	guint *strides;
};
typedef struct _QtecArrayInfo QtecArrayInfo;

/* Create a new empty wrapper. We force the user to define the element
 * size to ensure that memory handling can be consistently done.
 */
QtecArray *qtec_array_new(guint ele_size);

/* Creates a new array with the same elem_size, dimensions, strides and
 * endianness of the input array
 * Equivalent to calling: qtec_array_new_endian(), qtec_array_resize()
 * and qtec_array_set_dims_and_strides() in succession
 */
QtecArray *qtec_array_new_from_array(QtecArray *array);

/* Create a wrapper around pointed memory. The allocator should be
 * set depending on how the memory should be handled
 */
QtecArray *qtec_array_new_take(QtecMemory *mem, guint ele_size, guint elements);

/* Same as above but the user defines the endianness of the data
 */
QtecArray *qtec_array_new_endian(guint ele_size,
				 QtecArrayEndianness endianness);

/* Same as above but the user defines the endianness of the data
 */
QtecArray *qtec_array_new_take_endian(QtecMemory *mem, guint ele_size,
				      guint elements, QtecArrayEndianness end);

QtecArray *qtec_array_ref(QtecArray *array);
void qtec_array_unref(QtecArray *array);

/* sets the clear function for when data is removed */
void qtec_array_set_clear_func(QtecArray *array, GDestroyNotify clear_func);

/* Gets a raw pointer to the data. This is the beginning of the allocated
 * memory, so there may be padding before the actual data. This should primarily
 * be used when you need to e.g. pass the memory to a different library. For
 * specific index lookup use qtec_array_get_nd_element_pointer or
 * qtec_array_get_element_pointer
 */
gpointer qtec_array_get_data(QtecArray *array);

/* Converts a Multi dimentional array to a single dimention. Useful for resetting the array
 */
void qtec_array_flatten(QtecArray *arr);

/* Tries to set the dimensions of the array. The dimensions need to fit the size
 * of the array, otherwise this function returns false.
 */
gboolean qtec_array_set_dims_from_array(QtecArray *array, guint dims,
					guint32 dimarray[]);

/* Get a newely allocated info structure, containing information about the
 * array. Free when done using qtec_array_free_info
 */
gboolean qtec_array_get_info(QtecArray *arr, QtecArrayInfo **info);
void qtec_array_free_info(QtecArrayInfo *info);

/* Gets the nth dimention size. Will return 0 if dim >= maxdim */
guint qtec_array_get_nth_dim_size(const QtecArray *array, guint dim);
/* the number of bytes per element */
guint qtec_array_get_element_size(const QtecArray *array);
/* the number of elements in the array. */
guint qtec_array_get_elements(const QtecArray *array);
/* the number of dimensions of the array. */
guint qtec_array_get_nr_dims(const QtecArray *array);

/* Gets an element from the array. Checks that mem address is within
 * limits of mapped memory. Do note that there is no handling in terms of the
 * size of the pointed memory. It is the user that has to handle it correctly
 */
void *qtec_array_get_nd_element_pointer(QtecArray *array, guint pos_arr_no,
					guint posarray[]);
//when using the "raw" array position remember that things are stored
//in row-major order: y,x instead of x,y
void *qtec_array_get_element_pointer(QtecArray *array, guint pos);

/* Appends data to the array, reallocating to fit the new memory region
 */
gboolean qtec_array_append_vals(QtecArray *array, gconstpointer data,
				guint len);
/* removes an element from the index. The data following is shifted to keep the memory contugeous */
gboolean qtec_array_remove_index(QtecArray *array, guint idx);

/* Reallocates memory to fit x elements.
 */
gboolean qtec_array_resize(QtecArray *array, guint elements);

/* A combination of resize and set_dims_from_array. Does some
 * additional sanitizing and bundles common code at array initialization.
 */
gboolean qtec_array_resize_and_set_dims(QtecArray *arr, guint dim_no,
					guint32 dimarray[]);

/* Deep copy function. For shallow copies used ref/unref. If the allocator does not
 * provide a copy function, this returns NULL
 */
QtecArray *qtec_array_copy(QtecArray *array);

/* Attempts to create a string reprisentation of a mapped region of memory
 * it will try to cast memory for known regions, showing hex blocks
 */
gchar *qtec_array_serialize(QtecArray *array);

/* Changes the endianness of the array. If the array is already of the target
 * endianness, the function just returns. Otherwise a conversion is performed
 * for compatible element sizes: 16, 32 and 64 bit.
 */
void qtec_array_swap_endianness(QtecArray *array,
				QtecArrayEndianness endianness);

/* new API */
gboolean qtec_array_is_sparse(QtecArray *arr);

/* sets the dimensions and strides for the array, testing that
 * all alignments will work for the given settings. If any of the strides
 * are set to zero, they will be recalculated with the assumption that no
 * padding is being added for that dimension
 */
gboolean qtec_array_set_dims_and_strides(QtecArray *array, guint dimno,
					 guint32 dims[], guint32 strides[]);

/* Locks or unlocks the array resize operations, allowing to reuse an array in
 * a different fashion. This is useful especially for repurposing sparse arrays.
 *
 */
void qtec_array_set_resize_lock(QtecArray *arr, gboolean resizable);

/* Create a wrapper around pointed memory. Similar to qtec_array_new_take,
 * difference being that this can be used to define sparse memory.
 */
QtecArray *qtec_array_new_take_full(QtecMemory *mem, guint ele_size,
				    guint dimno, guint32 dims[],
				    guint32 strides[], guint offset,
				    QtecArrayEndianness endianness);

/* Retrieves the stride for the specified dimension of the array.
 */
gboolean qtec_array_get_stride(QtecArray *arr, guint pos, guint32 *stride);

/* Get the size of the offset, before valid data
 */
gsize qtec_array_get_offset(QtecArray *arr);
/* Get the size of the valid data region. This includes trailing padding
 */
gsize qtec_array_get_size(QtecArray *arr);
/* Get the entire data region of the buffer. This is offset + valid data +
 * padding
 */
gsize qtec_array_get_max_size(QtecArray *arr);

/* Retrieve the endianness of the array
 */
QtecArrayEndianness qtec_array_get_endianness(const QtecArray *array);

// Does the Array endianness matches the HOST endianness?
gboolean qtec_array_endianness_matches_host(const QtecArray *array);

/* Get a contiguous version of the input array. If the array is already
 * contiguous, we get a copy. The returned array needs to be unreffed
 */
QtecArray *qtec_array_sparse_to_contiguous(QtecArray *sparse_arr);

/* Iteratos */

/* Array memory iterator. retrieve one by one the contigeous memory chunks.
 * This is useful for operations that require to either read from or write
 * to memory that does not share the same layout as the array. The iterator
 * is valid as long as the array remains unchanged
 */
QtecIterator *qtec_array_chunk_memory_iterator_new(QtecArray *arr);

/* Array element iterator returns a pointer to the start of every element index.
 * it is up to the discrecion of the user to translate that pointer into
 * whatever the array reprisents.
 */
QtecIterator *qtec_array_element_iterator_new(QtecArray *arr);

G_END_DECLS

#endif
