/*
* @Author: Dimitrios Katsaros
* @Date:   2017-10-11 17:06:44
*/

/*
 * A basic debugging facility. Provides basic handles for
 * printing debug info and integration with other frameworks
 */

#ifndef __QTEC_LIB_CORE_DEBUG__
#define __QTEC_LIB_CORE_DEBUG__

#include <glib.h>
#include <qtec/core/debug.h>
#include "qtec/core/config.h"

G_BEGIN_DECLS
typedef enum {
	QTEC_DEBUG_LEVEL_NONE = 0,
	QTEC_DEBUG_LEVEL_ERROR = (1 << 0),
	QTEC_DEBUG_LEVEL_WARNING = (1 << 6),
	QTEC_DEBUG_LEVEL_INFO = (1 << 12),
	QTEC_DEBUG_LEVEL_DEBUG = (1 << 24),
	QTEC_DEBUG_LEVEL_DEEP_DEBUG = (1 << 30),

	QTEC_DEBUG_LEVEL_ALL = ~(QTEC_DEBUG_LEVEL_NONE),
} QtecDebugLevel;

typedef enum {
	QTEC_DEBUG_FLAGS_ALL_OFF = 0,
	QTEC_DEBUG_FLAGS_COLOR_OFF = (1 << 0),
} QtecDebugFlags;

typedef enum {
	QTEC_DEBUG_COLOR_NONE = 0,
	QTEC_DEBUG_COLOR_BLACK = 0, /* black is the default */
	QTEC_DEBUG_COLOR_RED,
	QTEC_DEBUG_COLOR_GREEN,
	QTEC_DEBUG_COLOR_YELLOW,
	QTEC_DEBUG_COLOR_BLUE,
	QTEC_DEBUG_COLOR_MAGENTA,
	QTEC_DEBUG_COLOR_CYAN,
	QTEC_DEBUG_COLOR_WHITE,

	QTEC_DEBUG_COLOR_BG_BLACK = (0 << 3),
	QTEC_DEBUG_COLOR_BG_RED = (1 << 3),
	QTEC_DEBUG_COLOR_BG_GREEN = (2 << 3),
	QTEC_DEBUG_COLOR_BG_YELLOW = (3 << 3),
	QTEC_DEBUG_COLOR_BG_BLUE = (4 << 3),
	QTEC_DEBUG_COLOR_BG_MAGENTA = (5 << 3),
	QTEC_DEBUG_COLOR_BG_CYAN = (6 << 3),
	QTEC_DEBUG_COLOR_BG_WHITE = (7 << 3),

	/* Other options */
	/* 0b000111 */
	QTEC_DEBUG_COLOR_FG_MASK = 0x7,
	/* 0b111000 */
	QTEC_DEBUG_COLOR_BG_MASK = 0x38,
	QTEC_DEBUG_COLOR_BOLD = (1 << 7),
	QTEC_DEBUG_COLOR_UNDERLINE = (1 << 8),
	QTEC_DEBUG_COLOR_REVERSED = (1 << 9),
	/* overrides all other settings */
	QTEC_DEBUG_COLOR_RESET = ~(QTEC_DEBUG_COLOR_NONE),
} QtecColorFlags;

extern QtecDebugLevel _active_debug;

void qtec_debug_set_active_debug(QtecDebugLevel level);
void qtec_debug_set_flags(QtecDebugFlags level);
void qtec_debug_reset_flags(void);

typedef void (*QtecLogFunction)(QtecDebugLevel level, const gchar *file,
				const gchar *func, gint line,
				const gchar *message, gpointer user_data);

void qtec_debug_default_log(QtecDebugLevel level, const gchar *file,
			    const gchar *func, gint line, const gchar *message,
			    gpointer user_data);
void qtec_debug_add_logging_function(QtecLogFunction func, gpointer data,
				     GDestroyNotify notify);
void qtec_debug_remove_logging_function(QtecLogFunction func);

QtecColorFlags qtec_debug_get_level_color_flags(QtecDebugLevel level);
const gchar *qtec_debug_create_debug_color(QtecColorFlags flags);
const gchar *qtec_debug_level_get_name(QtecDebugLevel level);
void qtec_debug_set_logging_level(QtecDebugLevel lvl, const gchar *name,
				  QtecColorFlags color);
void qtec_debug_log_valist(QtecDebugLevel level, const gchar *file,
			   const gchar *func, gint line, const gchar *format,
			   va_list args);
void qtec_debug_log(QtecDebugLevel level, const gchar *file, const gchar *func,
		    gint line, const gchar *format, ...);

static inline void
qtec_debug_filter_active_log_valist(QtecDebugLevel level, const gchar *file,
				    const gchar *func, gint line,
				    const gchar *format, va_list args)
{
	if (G_UNLIKELY(level & _active_debug))
		qtec_debug_log_valist(level, file, func, line, format, args);
}

static inline void qtec_debug_filter_active_log(QtecDebugLevel level,
						const gchar *file,
						const gchar *func, gint line,
						const gchar *format, ...)
{
	if (G_UNLIKELY(level & _active_debug)) {
		va_list vlist;

		va_start(vlist, format);
		qtec_debug_log_valist(level, file, func, line, format, vlist);
		va_end(vlist);
	}
}

/* the macros for printing category info */
#ifndef QTEC_DISABLE_DEBUG

#define QTEC_ERROR(...)                                                 \
	(qtec_debug_filter_active_log(QTEC_DEBUG_LEVEL_ERROR, __FILE__, \
				      __func__, __LINE__, __VA_ARGS__))
#define QTEC_WARNING(...)                                                 \
	(qtec_debug_filter_active_log(QTEC_DEBUG_LEVEL_WARNING, __FILE__, \
				      __func__, __LINE__, __VA_ARGS__))
#define QTEC_INFO(...)                                                 \
	(qtec_debug_filter_active_log(QTEC_DEBUG_LEVEL_INFO, __FILE__, \
				      __func__, __LINE__, __VA_ARGS__))
#define QTEC_DEBUG(...)                                                 \
	(qtec_debug_filter_active_log(QTEC_DEBUG_LEVEL_DEBUG, __FILE__, \
				      __func__, __LINE__, __VA_ARGS__))
#define QTEC_DEEP_DEBUG(...)                                                 \
	(qtec_debug_filter_active_log(QTEC_DEBUG_LEVEL_DEEP_DEBUG, __FILE__, \
				      __func__, __LINE__, __VA_ARGS__))

#else

#define QTEC_ERROR(...) \
	{               \
	}
#define QTEC_WARNING(...) \
	{                 \
	}
#define QTEC_INFO(...) \
	{              \
	}
#define QTEC_DEBUG(...) \
	{               \
	}

#endif

G_END_DECLS
#endif /* __QTEC_LIB_CORE_DEBUG__ */
