/*
 * @Author: Dimitrios Katsaros
 * @Date:   2018-06-04
 */

/* A memory wrapper and a set of allocators for constructing wrapped memory.
 * This is useful for components like QtecArray, that need to be capable of
 * handling multiple types of memory
 */

#ifndef __QTEC_LIB_ALLOCATOR_H__
#define __QTEC_LIB_ALLOCATOR_H__

#include <glib-object.h>

G_BEGIN_DECLS
#define QTEC_ALLOCATOR_MALLOC "QtecMallocAllocator"
#define QTEC_ALLOCATOR_SLICE "QtecSliceAllocator"

#define QTEC_MEMORY_CAST(mem) ((QtecMemory *)(mem))
#define QTEC_ALLOCATOR_TYPE (qtec_allocator_get_type())
#define QTEC_ALLOCATOR_CLASS(cls)                            \
	(G_TYPE_CHECK_CLASS_CAST((cls), QTEC_ALLOCATOR_TYPE, \
				 QtecAllocatorClass))
#define QTEC_ALLOCATOR_GET_CLASS(obj)                          \
	(G_TYPE_INSTANCE_GET_CLASS((obj), QTEC_ALLOCATOR_TYPE, \
				   QtecAllocatorClass))

typedef struct _QtecMemory QtecMemory;
typedef struct _QtecAllocator QtecAllocator;
typedef struct _QtecAllocatorClass QtecAllocatorClass;

/* An extended version of GDestroyNotify. You can cast GDestoryNotify compatible
 * pointers into this, like g_object_unref
 */
typedef void (*QtecMemDestroyNotify)(gpointer user_data, QtecMemory *mem);

/* These structs are public so they can be extenced by new memory/allocator
 * implementations
 */
struct _QtecMemory {
	QtecAllocator *allocator;
	gsize size;
};

struct _QtecAllocatorClass {
	GObjectClass parent;

	QtecMemory *(*alloc)(QtecAllocator *tmp, gsize size);
	void (*dealloc)(QtecMemory *mem);
	void (*realloc)(QtecMemory *mem, gsize size);
	void *(*map)(QtecMemory *mem);
	QtecMemory *(*copy)(QtecMemory *mem, gsize size);
};

GType qtec_allocator_get_type(void);

/* QtecAllocator */
/* Registers a new allocator that can be later retrieved. Takes ownership of
 * allocator
 */
gboolean qtec_allocator_register(const gchar *name, QtecAllocator *allocator);

/* Tries to get a previously registered allocator. fails if not found */
gboolean qtec_allocator_get(const gchar *name, QtecAllocator **allocator);

/* Gets the default allocator. */
QtecAllocator *qtec_allocator_get_default();

/* Tries to get a previously registered allocator. If we fail, the default
 * allocator will be returned
 */
QtecAllocator *qtec_allocator_try_get(const gchar *name);

/* Allocates a piece of memory according to the allocator. If no allocator
 * is specified it will use the default allocator
 */
QtecMemory *qtec_allocator_alloc(QtecAllocator *allocator, gsize size);

/* QtecMemory */
/* Wraps a pointer in qtec memory. you can use the notification to cleanup any
 * associated data.
 */
QtecMemory *qtec_memory_new_wrapped(gpointer data, gsize size,
				    QtecMemDestroyNotify cleanup,
				    gpointer user_data);

/* Initializes a memory slice with all relevant information. This
 * should only really be used when creating your own allocator and are
 * initializing a new piece of memory.
 */
void qtec_memory_init(QtecMemory *qmem, QtecAllocator *allocator, gsize size);
/* Frees memory. It uses the allocator that made it to decide how to release
 * the memory
 */
void qtec_memory_free(QtecMemory *qmem);

/* Tries to get a pointer to the associated memory. This may fail depending
 * on the way the underlaying memory is handled.
 */
gboolean qtec_memory_map_data(QtecMemory *qmem, void **ptr);

/* Reallocates the memory to the given size
 */
gboolean qtec_memory_realloc(QtecMemory *qmem, gsize size);
/* Creates a copy of the memory region. you can specify the size of the new
 * target. Setting the target size to zero will do a complete copy
 */
gboolean qtec_memory_copy(QtecMemory *mem, QtecMemory **copy, gsize size);

/* Get the total size of the data region
 */
gsize qtec_memory_get_size(QtecMemory *mem);

G_END_DECLS
#endif
