/*
 * @Author: Dimitrios Katsaros
 * @Date:   2017-11-20
 */

#ifndef __QTEC_GST_LINK_H__
#define __QTEC_GST_LINK_H__

#include <gst/gst.h>

G_BEGIN_DECLS
/* the following functions are based on camerabingeneral.h in gst-plugins-bad
 * under gst/camerabin2:
 * https://cgit.freedesktop.org/gstreamer/gst-plugins-bad
 * gst/camerabin2/camerabingeneral.h
 */
/* Helper function for link elements in a bin.
 * The srcpad has to be a pad of an element in bin or NULL.
 * If srcpad is NULL the element is only added.
 * New srcpad is the returned srcpad of the targetele. It can be NULL.
 * the filtercaps parameter can be used to filter which pad is chosen as the
 * target sink. This is useful when multiple sink pads are compatible.
 */
gboolean qtec_gst_link_add_ele_and_link(GstBin *bin, GstElement *target_ele,
					GstPad *srcpad, GstCaps *filtercaps,
					GstPad **new_srcpad, GError **error);

/* function for targeted link of two elements. Creates the ele and links it
 * together with the given pad, returning the srcpad of the created element.
 * If there is no srcpad defined, the new element is just added to the bin.
 * The return pad can be NULL
 */
GstElement *qtec_gst_link_create_typed_elem_and_link(GstBin *bin, GType eletype,
						     const gchar *instance_name,
						     GstPad *srcpad,
						     GstPad **ret_pad,
						     GError **error);

/* function for targeted link of two elements. Creates the ele and links it
 * together with the given pad, returning the srcpad of the created element.
 * If there is no srcpad defined, the new element is just added to the bin.
 * The return pad can be NULL
 */
GstElement *qtec_gst_link_create_named_elem_and_link(
	GstBin *bin, const gchar *elem_name, const gchar *instance_name,
	GstPad *srcpad, GstPad **ret_pad, GError **error);
/* end of camerabingeneral.h based functions */

/* A simple function for getting the first unlinked pad in a given direction
 * on an element
 */

GstPad *qtec_gst_link_find_unlinked_pad(GstElement *element,
					GstPadDirection direction);

/*
 * function for retrieving a request pad from an element and link it to a
 * target pad. Checks are performed so the 2 pads are linked correctly. can
 * request a src to link to a sink or vice versa
 */
gboolean qtec_gst_link_request_pad_and_link(GstElement *reqpad_souce,
					    gchar *template_name,
					    GstPad *target_pad,
					    GstPad **ret_pad, GError **error);

struct QtecGstLinkInfo;

/* Function pointer for callback after delayed link has succeeded */
typedef void (*PostLinkHandler)(gpointer userdata);

/* adds callback for when a delayed link is completed */
void qtec_gst_link_add_post_link_handler(struct QtecGstLinkInfo *info,
					 PostLinkHandler handler,
					 gpointer usrptr);

/* helper function for building pipelines out of elements with maybe pads
 * Params: ele - An element that will create a maybe pad.
 *         pad - The pad to try and link with
 */
struct QtecGstLinkInfo *qtec_gst_link_new_delayed_link(GstElement *ele,
						       GstPad *pad);

/* Look for a specific pad name to link to. Same as above with added maybe
 * pad name that we wish to connect to.
 */
struct QtecGstLinkInfo *
qtec_gst_link_new_delayed_link_named(GstElement *ele, gchar *new_pad_name,
				     GstPad *pad);

/* Variant of 'new_delayed_link' function. This version takes an element with a
 * request pad and tries to link the request pad with the dynamic pads generated
 * by the sometimes element.
 * Params: sometimes_ele - An element that will create maybe pads.
 *         request_ele   - The element from which to retrieve a request pad from
 *         template_name - The template of the request pad to retrieve
 *         retpad        - The request pad that was found. This is useful for
 *                         setting pad parameters. Does not need to be
 *                         dereferenced
 */

struct QtecGstLinkInfo *qtec_gst_link_new_delayed_link_to_request_pad(
	GstElement *sometimes_ele, GstElement *request_ele,
	const gchar *template_name, GstPad **retpad);

G_END_DECLS
#endif /* __QTEC_GST_LINK_H__ */
