/*
 * @Author: Dimitrios Katsaros
 * @Date:   2017-11-23
 */

#ifndef __QTEC_LIB_IO_PAM_H__
#define __QTEC_LIB_IO_PAM_H__

/* A wrapper around a QtecArray for saving images and (theoretically) any data
 * that is arranged as a 2d rectangular array using the netpbm and PAM formats.
 * Note that both netpbm and PAM Expect the data to be big endian. The QtecArray
 * will swap to the correct endianness for operations.
 */
#include <qtec/io/pam/info.h>
#include <qtec/core/array.h>
#include <glib.h>
#include <gio/gio.h>

G_BEGIN_DECLS

typedef struct _QtecIOPam QtecIOPam;

QtecIOPam *qtec_io_pam_ref(QtecIOPam *pam);
void qtec_io_pam_unref(QtecIOPam *pam);
QtecIOPam *qtec_io_pam_copy(QtecIOPam *pam);
/* Creates pam wrapper from array. Fields are set based on array layout.
 * Defaut encoding to binary.
 */
QtecIOPam *qtec_io_pam_new(guint width, guint height, guint depth,
			   guint ele_size, QtecIOPamType type);
//expects the array to be in the form: height, width, depth (row-major)
QtecIOPam *qtec_io_pam_new_from_array(QtecArray *array, QtecIOPamType type);
QtecIOPam *qtec_io_pam_read_from_stream(GInputStream *stream,
					GCancellable *cancellable,
					GError **error);
gboolean qtec_io_pam_write_to_stream(QtecIOPam *pam, GOutputStream *stream,
				     GCancellable *cancellable, GError **error);
QtecIOPam *qtec_io_pam_from_file(const gchar *fileloc,
				 GCancellable *cancellable, GError **error);
gboolean qtec_io_pam_write_to_file(QtecIOPam *pam, const gchar *filepath,
				   GCancellable *cancellable, GError **error);

/* helper function for checking if tuple type is part of pam */
gboolean qtec_io_pam_check_for_tuple_type(QtecIOPam *pam,
					  const gchar *tuple_type);
/* Attempts to add a type type to the pam. Tuple types are only valid for PAM
 * images. The function checks that the string is valid.
 */
gboolean qtec_io_pam_add_tuple_type(QtecIOPam *pam, const gchar *type_type);
/* Retrieves the list of set tuple types
 */
GList *qtec_io_pam_get_tuple_types(QtecIOPam *pam);
/* Clears all the tuple types from the wrapper */
void qtec_io_pam_clear_tuple_types(QtecIOPam *pam);
/* Attempts to set the encoding for the image. May return false if the
 * type does not support the selected encoding. I.e. PAM images can only
 * be binary
 */
gboolean qtec_io_pam_set_encoding(QtecIOPam *pam, QtecIOPamEncoding enc);

//when using the "raw" array position remember that things are stored
//in row-major order: y,x,channel instead of x,y,channel
guint qtec_io_pam_get_value(QtecIOPam *pam, guint pos, GError **error);
gboolean qtec_io_pam_set_value(QtecIOPam *pam, guint val, guint pos,
			       GError **error);
guint qtec_io_pam_get_nd_value(QtecIOPam *pam, guint x, guint y, guint plane,
			       GError **error);
gboolean qtec_io_pam_set_nd_value(QtecIOPam *pam, guint val, guint x, guint y,
				  guint plane, GError **error);
/* retrieve raster info. Any unwanted fields can be ignored by passing NULL */
void qtec_io_pam_get_raster_params(QtecIOPam *pam, QtecIOPamType *type,
				   QtecIOPamEncoding *enc, guint *x, guint *y,
				   guint *depth, guint *maxval);
/* Gets a reference to the underlaying qtec array. Make sure you are careful
 * with how you handle the array. Performing operations on the data will have
 * an immediate impact on the wrapper and may cause undefined behaviour
 */
QtecArray *qtec_io_pam_get_array_ref(QtecIOPam *pam);
/* Gets a copy of the memory contained in the pam wrapper. Unlike
 * qtec_io_pam_get_array_refs, this array is a deep copy and can be freely
 * manipulated.
 */
QtecArray *qtec_io_pam_get_copy(QtecIOPam *pam);

/* info functions */
const gchar *qtec_io_pam_type_to_string(QtecIOPamType t);
const gchar *qtec_io_pam_encoding_to_string(QtecIOPamEncoding e);
guint qtec_io_pam_type_to_number(QtecIOPamType type, QtecIOPamEncoding enc);
G_END_DECLS
#endif /* __QTEC_LIB_IO_PAM_H__ */
