/*
* @Author: Dimitrios Katsaros
* @Date:   2017-05-04 11:43:45
*/

#ifndef __QTEC_LIB_V4L2_CORE_H__
#define __QTEC_LIB_V4L2_CORE_H__

#include <glib.h>
#include <linux/videodev2_qtec.h>

G_BEGIN_DECLS
#define V4L2_DEV_DISABLED_FLAG                               \
	(V4L2_CTRL_FLAG_DISABLED | V4L2_CTRL_FLAG_INACTIVE | \
	 V4L2_CTRL_FLAG_GRABBED)

struct _QtecV4l2CtrlClass {
	guint ctrlclassid;
	guint ctrlno;
	GList *ctrlids;
};

typedef struct _QtecV4l2Dev QtecV4l2Dev;
typedef struct _QtecV4l2CtrlClass QtecV4l2CtrlClass;

/* device handling */
void qtec_v4l2_device_unref(QtecV4l2Dev *dev);

/* Locking functions
 */
void qtec_v4l2_lock(QtecV4l2Dev *dev);
void qtec_v4l2_unlock(QtecV4l2Dev *dev);

/* Can use a printf pattern, eg '/dev/video%d', works only with %d and %u,
 * found name can be NULL. If non null, a new string is allocated. Free with g_free
 */
gboolean qtec_v4l2_find_device(const gchar *filename, gchar **found_name);
gboolean qtec_v4l2_open_device(const gchar *loc, QtecV4l2Dev **dev);
gboolean qtec_v4l2_find_and_open_device(const gchar *file_name,
					QtecV4l2Dev **dev);

//search in all the defined paths
//remember to g_free(devloc)
gboolean qtec_v4l2_video_dev_search(gchar **devloc);
gboolean qtec_v4l2_video_dev_search_and_open(QtecV4l2Dev **dev);

/* basic functions */
gboolean qtec_v4l2_get_fmt(QtecV4l2Dev *dev, __u32 type,
			   struct v4l2_format *fmt);
gboolean qtec_v4l2_get_caps(QtecV4l2Dev *dev, struct v4l2_capability *caps);

gboolean qtec_v4l2_enum_fmt(QtecV4l2Dev *dev, struct v4l2_fmtdesc *fmt);
gboolean qtec_v4l2_enum_frame_sizes(QtecV4l2Dev *dev,
				    struct v4l2_frmsizeenum *fs);
gboolean qtec_v4l2_enum_frame_intervals(QtecV4l2Dev *dev,
					struct v4l2_frmivalenum *fi);
gboolean qtec_v4l2_get_param(QtecV4l2Dev *dev,
			     struct v4l2_streamparm *streamaparm);
gboolean qtec_v4l2_set_param(QtecV4l2Dev *dev,
			     struct v4l2_streamparm *streamaparm);

/* utility funcs */
gint qtec_v4l2_get_control_id(QtecV4l2Dev *dev, const gchar *control_name);
QtecV4l2CtrlClass *qtec_v4l2_get_ctrl_class(QtecV4l2Dev *dev, guint id);
gboolean qtec_v4l2_lookup_qctrl_by_name(QtecV4l2Dev *dev, const gchar *name,
					struct v4l2_query_ext_ctrl **retqry);
gboolean qtec_v4l2_lookup_qctrl_by_id(QtecV4l2Dev *dev, guint id,
				      struct v4l2_query_ext_ctrl **retqry);

/* extended controls */
/* allocation/deallocaton functions
 */
void qtec_v4l2_ext_control_free_payload(struct v4l2_query_ext_ctrl *qctrl,
					struct v4l2_ext_control *ctrl);
void qtec_v4l2_ext_control_allocate_payload(struct v4l2_query_ext_ctrl *qctrl,
					    struct v4l2_ext_control *ctrl);
void qtec_v4l2_ext_control_copy_payload(struct v4l2_query_ext_ctrl *qctrl,
					struct v4l2_ext_control *srcctrl,
					struct v4l2_ext_control *targetctrl);
void qtec_v4l2_free_ext_controls(QtecV4l2Dev *dev,
				 struct v4l2_ext_controls *ctrls);
struct v4l2_ext_controls *
qtec_v4l2_alloc_ext_controls(struct v4l2_query_ext_ctrl *qctrlarr, guint count);
struct v4l2_ext_controls *
qtec_v4l2_copy_ext_controls(QtecV4l2Dev *dev,
			    struct v4l2_ext_controls *srcctrls);

gboolean qtec_v4l2_set_ext_ctrl_default(QtecV4l2Dev *dev,
					struct v4l2_ext_control *ctrl);
gboolean qtec_v4l2_set_ext_controls(QtecV4l2Dev *dev,
				    struct v4l2_ext_controls *controls);
/* Function for receiving a single v4l2 control. The controls structure is allocated and should be freed
 * using qtec_v4l2_free_ext_controls
 */
gboolean qtec_v4l2_get_ext_control(QtecV4l2Dev *dev, int control_id,
				   struct v4l2_ext_controls **ret_controls);
/* Function for receiving a single v4l2 control. As opposed to the previous one, the controls structure needs
 * to be allocated by the user.
 */
gboolean
qtec_v4l2_get_ext_control_using_struct(QtecV4l2Dev *dev, int control_id,
				       struct v4l2_ext_controls *controls);
/* Function for receiving v4l2 controls. This is a pure ioctl call so the user is in charge of setting up
 * all the fields and preallocating the required space to fit all the requested controls
 */
gboolean qtec_v4l2_get_ext_controls(QtecV4l2Dev *dev,
				    struct v4l2_ext_controls *controls);
gboolean qtec_v4l2_query_ext_control(QtecV4l2Dev *dev,
				     struct v4l2_query_ext_ctrl *qctrl,
				     gboolean next_ctrl);

/* events */
void qtec_v4l2_subscribe_event(QtecV4l2Dev *dev, int id,
			       gboolean allow_feedback);
void qtec_v4l2_unsubscribe_event(QtecV4l2Dev *dev, int id);
void qtec_v4l2_subscribe_ctrl_events(QtecV4l2Dev *dev, gboolean allow_feedback);
void qtec_v4l2_unsubscribe_all_events(QtecV4l2Dev *dev);
gboolean qtec_v4l2_get_event(QtecV4l2Dev *dev, struct v4l2_event *event);
gint qtec_v4l2_poll_for_event(QtecV4l2Dev *dev, gint timeout);

/* Additional v4l2 operations */
gboolean qtec_v4l2_query_menu(QtecV4l2Dev *dev, struct v4l2_querymenu *qmenu);
gboolean qtec_v4l2_get_selection(QtecV4l2Dev *dev, struct v4l2_selection *sel);
gboolean qtec_v4l2_set_fmt_full(QtecV4l2Dev *dev, struct v4l2_format *format,
				gboolean try_only);

//"new API" for finding control "classes"
//based on which FPGA core they belong to
//remember to g_free(str);
gchar *qtec_v4l2_get_ctrl_fpga_core(guint ctrl_id);

G_END_DECLS
#endif
